import java.util.*;
import java.io.*;
import java.util.concurrent.*;


class Executor extends Thread {

    public static final int k = 5;
    public static final int n = (1 << k)+2;
    public static final double h = 1.0/(n/2.0-1.0);
    public static final double dt = 1e-4;
    public static final double tmax = 100*dt;
    public static final int tsteps = (int)(tmax/dt);

    public static ArrayList<ArrayList<Production>> productionTree;

    public static Vertex makeTree(int k) throws InterruptedException, BrokenBarrierException {
        Vertex S = new Vertex(null, null, null, "S");
        productionTree = new ArrayList<ArrayList<Production>>();
        ArrayList<Production> currentProductions = new ArrayList<Production>();
        productionTree.add(currentProductions);
        
        CyclicBarrier cb = new CyclicBarrier(2);
        Production p1 = new P1(S, cb);
        p1.start();
        currentProductions.add(p1);
        cb.await();
        
        ArrayList<Production> newProductions;
        for(int i = 0; i < k-1; ++i) {
            newProductions = new ArrayList<Production>();
            cb = new CyclicBarrier(currentProductions.size()*2+1);
            for(Production pp : currentProductions) {
                Production p2a = new P2(pp.m_vertex.m_left,cb);
                Production p2b = new P2(pp.m_vertex.m_right,cb);
                newProductions.add(p2a);
                newProductions.add(p2b);
            }
            for (Production p : newProductions)
                p.start();
            cb.await();
            currentProductions = newProductions;
            productionTree.add(currentProductions);
        }   
        
        newProductions = new ArrayList<Production>();
        cb = new CyclicBarrier(currentProductions.size()*2+1);
        for(Production pp : currentProductions) {
            Production p3a = new P3(pp.m_vertex.m_left,cb);
            Production p3b = new P3(pp.m_vertex.m_right,cb);
            newProductions.add(p3a);
            newProductions.add(p3b);
        }
        for (Production p : newProductions)
            p.start();
        cb.await();
        productionTree.add(newProductions);
        
        return S;
    }
    
    public double[] solve(double [] previous) throws InterruptedException, BrokenBarrierException {
		double[] result = new double[n];
		
		Vertex S = makeTree(k);
		
        //LEAFS
        List<Production> made = new ArrayList<Production>();		
		ArrayList<Production> leafs = productionTree.get(productionTree.size()-1);
		
        CyclicBarrier cb = new CyclicBarrier(leafs.size()+1);
		made.add(new A1(leafs.get(0).m_vertex, cb, previous[0], previous[2]));
		made.add(new A00(leafs.get(1).m_vertex, cb));
		for(int i = 2; i < leafs.size()-2; i+=2)
		{
		    made.add(new A(leafs.get(i).m_vertex, cb, previous[i], previous[i+2]));
		    made.add(new A00(leafs.get(i+1).m_vertex, cb));
		}
		made.add(new AN(leafs.get(leafs.size()-2).m_vertex, cb, previous[n-4], previous[n-2]));
		made.add(new A01(leafs.get(leafs.size()-1).m_vertex, cb));
		for(Production p : made)
		    p.start();
        cb.await();
		
		for (Production p : made)
		    p.m_vertex.dumpMatrix();
		    
		// SECOND
		
		made = new ArrayList<Production>();		
		ArrayList<Production> second = productionTree.get(productionTree.size()-2);
		
        cb = new CyclicBarrier(second.size()+1);

		for(Production p : second)
		    made.add(new A2(p.m_vertex, cb));
		for(Production p : made)
		    p.start();
		    
        cb.await();
		for (Production p : made)
		    p.m_vertex.dumpMatrix();
		
		// VERTEXES
		for(int j = productionTree.size()-3; j >= 1; --j)
		{
		    List<Production> currentLevel = productionTree.get(j);
		    System.out.println("LEVEL");
		    System.out.println(currentLevel.size());    
		    made = new ArrayList<Production>();
		    cb = new CyclicBarrier(currentLevel.size()+1);
		    for (Production p : currentLevel)
		        made.add(new A3(p.m_vertex, cb));
		    for (Production p : made)
		        p.start();
		    cb.await();
		    
   	        for (Production p : made) {
    		    p.m_vertex.dumpMatrix();
    		}
		    
   		    made = new ArrayList<Production>();
		    for (Production p : currentLevel)
		        made.add(new E3(p.m_vertex, cb));
		    for (Production p : made)
		        p.start();
		    cb.await();
		    
		    for (Production p : made)
    		    p.m_vertex.dumpMatrix();
		    
		    System.out.println("NEXT LEVEL");
		}
		cb = new CyclicBarrier(2);
		Production root = productionTree.get(0).get(0);
		(new Aroot(root.m_vertex, cb)).start();
		cb.await();
		root.m_vertex.dumpMatrix();
		(new Eroot(root.m_vertex, cb)).start();
		cb.await();
		root.m_vertex.dumpMatrix();
		(new BS1(root.m_vertex, cb)).start();
		cb.await();
		root.m_vertex.dumpMatrix();
		
		for (int j = 1; j < productionTree.size()-2; ++j)
		{
		    List<Production> currentLevel = productionTree.get(j);
		    made = new ArrayList<Production>();
		    
		    cb = new CyclicBarrier(currentLevel.size()+1);
		    for (Production p : currentLevel)
		        made.add(new BS1(p.m_vertex, cb));
		    for (Production p : made)
		        p.start();
		    cb.await();
		    
		    for (Production p : made)
		        p.m_vertex.dumpMatrix();
		}
		
		second = productionTree.get(productionTree.size()-2);
		made = new ArrayList<Production>();
		cb = new CyclicBarrier(second.size()+1);
	    for (Production p : second)
	        made.add(new BS2(p.m_vertex, cb));
	    for (Production p : made)
	        p.start();
	    cb.await();
	    
	    for (Production p : made)
	        p.m_vertex.dumpMatrix();
		
		
		for (Production l : leafs) 
		    l.m_vertex.dumpMatrix();    
		    
		// WRITE TO OUTPUT
		
		int i = 0;
		for (Production l : leafs) {
		    result[i++] = l.m_vertex.m_x[2];
		}
		result[n-2] = leafs.get(leafs.size()-1).m_vertex.m_x[3];
		result[n-1] = leafs.get(leafs.size()-1).m_vertex.m_x[4];
		
		return result;
    }


    
	public synchronized void run() {
    
        try {	         
            double[][] u = new double[tsteps+1][n];
            for (int i = 0; i < n; ++i)
                u[0][i] = 0;
                
            System.out.println(tsteps);
                    System.out.println(tmax/dt);
            for(int t = 1; t <= tsteps; t++) {
                System.out.println("STEP");
                A.previousSolution=u[t-1];
                u[t] = solve(u[t-1]);
            }
        
            PrintWriter f = new PrintWriter("results.txt");
            double h = 1.0/n;
		    
		    for (int i = 0; i <= tsteps; ++i) {
		        f.print("\"t=");
		        f.print(dt*i);
		        f.println("\"");
		        double x = 0.0;
		        for(int j = 0; j < n; ++j) {
		            f.print(x);
		            f.print(" ");
		            f.println(u[i][j]);
		            x += h;
		        }
		        f.println();
   		        f.println();
		    }
		    f.close();
	        
            Runtime r = Runtime.getRuntime();
            Process p = r.exec("echo 'plot \"results.txt\"; pause -1' | gnuplot");
            p.waitFor();
        } catch(Exception ex) {
            ex.printStackTrace();
        }
	}
}









